package com.ibm.sysmgt.raidmgr.wizard.scan;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.awt.*;
import java.util.Vector;
import java.util.Enumeration;

import com.ibm.sysmgt.raidmgr.util.*;
import com.ibm.sysmgt.storage.api.StorRet;
import com.ibm.sysmgt.raidmgr.dataproc.config.*;
import com.ibm.sysmgt.raidmgr.dataproc.util.DeviceID;
import com.ibm.sysmgt.raidmgr.images.*;
import com.ibm.sysmgt.raidmgr.mgtGUI.Launch;
import com.ibm.sysmgt.raidmgr.mgtGUI.GUIDataProc;
import com.ibm.sysmgt.raidmgr.mgtGUI.dialogs.OpFailedDialog;


public class ScanDrivesWizard extends JPanel implements Runnable, ActionListener
{
   private Launch launch;   
   private boolean stop;   
   private Adapter adapter;
   private Channel channel;
   private int adapterID;
   private int channelID;
   private Thread th;
   private GUIDataProc  dp;
   private int numRemoved = 0;
   private int numAdded   = 0;

   private JProgressBar progress;
   private JButton cancelButton, doneButton;
   private JLabel       label, lrem, ladd;

   public ScanDrivesWizard(Launch launch, Adapter adapter, Channel channel)
   {                  
      super(new GridBagLayout());
      this.launch = launch;

      this.adapter = adapter;
      this.channel = channel;

      adapterID = adapter.getID();
      channelID = (channel == null) ? 99 : channel.getID();
      
      dp = (GUIDataProc) launch.getDP();

      setupGUI();
   }

   public void run()
   {
      // Initialize the counts of how many we found added and removed
      numRemoved     = 0;
      numAdded       = 0;

      StorRet ret = new StorRet();

      Vector vPhyDev = new Vector(60);
      Adapter ad     = null;
      int numChan    = 0;
      
      RaidSystem system = dp.getConfigAll();

      if (system == null)
      {
         Object[] arg = { new Integer(adapter.getAdjustedID())};
         OpFailedDialog fd = new OpFailedDialog(launch, "opFailedController1", arg, "opFailedController2");
         fd.show();
         
         // Fire cancel action
         cancelButton.doClick();
         return;
      }
      for (Enumeration e = system.enumerateAdapters() ; e.hasMoreElements() ;)
      {
         if (endScan(stop)) return;

         ad = (Adapter) e.nextElement();
         if (ad.getID() == adapterID)
         {
            for (Enumeration e2 = ad.getPhysicalDeviceCollection(null).elements(); e2.hasMoreElements();)
               vPhyDev.addElement(e2.nextElement());

            numChan = ad.getChannelCount();
            break;
         }
      }

      // Need larger counts to reflect the enclosure un-discovery process
      int totalnum       = (channelID == 99) ? (vPhyDev.size() + 1) : (ad.getLimit(Adapter.MAX_SCSI_ID) + 2);
      int index          = 0;
      PhysicalDevice dev = null;
      Adapter ad1;

      for (Enumeration e = vPhyDev.elements() ; e.hasMoreElements() ;)
      {
         if (endScan(stop)) return;

         dev = (PhysicalDevice) e.nextElement();

         if (channelID == 99 || channelID == dev.getChannelID())
         {
            if (dev.getState() == Constants.RDY)
            {
               ret = dp.setPhysDevState(new DeviceID(adapterID, dev.getChannelID(), dev.getDeviceID()), (short) Constants.EMP);
               if (ret.iReturnCode == 0)
               {
                  ret = dp.setPhysDevState(new DeviceID(adapterID, dev.getChannelID(), dev.getDeviceID()), (short) Constants.RDY);
                  if (ret.iReturnCode != 0)
                  {
                     numRemoved++;
                     Object[] args = {new Integer(numRemoved)};
                     String s2 = JCRMUtil.makeNLSString( "scanNumRemoved", args);
                     lrem.setText(s2);
                  }
               }
               // Scan the passive controller if this controller is part of a failover pair
               if ((ad instanceof ServeRaidAdapter) && (((ServeRaidAdapter)ad).getPairCount() != 0))
               {
                  for (Enumeration e1 = system.enumerateAdapters(); e1.hasMoreElements();)
                  {
                     ad1 = (Adapter) e1.nextElement();
                     if ((ad1 instanceof ServeRaidAdapter) && 
                         ((ServeRaidAdapter)ad1).getClusterPartnerID().equals(((ServeRaidAdapter)ad).getHostID()))
                     {
                        dp.setPhysDevState(new DeviceID(ad1.getID(), dev.getChannelID(), dev.getDeviceID()), (short) Constants.EMP);
                        dp.setPhysDevState(new DeviceID(ad1.getID(), dev.getChannelID(), dev.getDeviceID()), (short) Constants.RDY);
                        break;
                     }
                  }
               }
            }
            else if (dev.getState() == Constants.EMP)
            {
               //Try to set empty devices to ready; this will detect added hard drives.
               ret = dp.setPhysDevState(new DeviceID(adapterID, dev.getChannelID(), dev.getDeviceID()), (short) Constants.RDY);
               if (ret.iReturnCode == 0)
               {
                  numAdded++;
                  Object[] args = {new Integer(numAdded)};
                  String s1 = JCRMUtil.makeNLSString( "scanNumFound", args);
                  ladd.setText(s1);
               }
               else
               {
                  //Not a hard drive, so look for a standby device, such as a SAFTE chip
                  dp.setPhysDevState(new DeviceID(adapter.getID(), dev.getChannelID(), dev.getDeviceID()), (short) Constants.SBY); // detect enclosures
               }

               // Scan the passive controller if this controller is part of a failover pair
               if ((ad instanceof ServeRaidAdapter) && (((ServeRaidAdapter)ad).getPairCount() != 0))
               {
                  for (Enumeration e1 = system.enumerateAdapters() ; e1.hasMoreElements();)
                  {
                     ad1 = (Adapter) e1.nextElement();
                     if ((ad1 instanceof ServeRaidAdapter) && 
                         ((ServeRaidAdapter)ad1).getClusterPartnerID().equals(((ServeRaidAdapter)ad).getHostID()))
                     {
                        dp.setPhysDevState(new DeviceID(ad1.getID(), dev.getChannelID(), dev.getDeviceID()), (short) Constants.RDY);
                        break;
                     }
                  }
               }
            }
            index++;
         }
         int val = index*100/totalnum;
         progress.setValue(val);
         if (endScan(stop)) return;
      }

      // Now look for enclosures to remove on otherwise empty channels
      removeEnclosureDevices();
      

      //This call instructs the firmware to validate its configuration, which associates drives with enclosures.  
      //If this call is not made, then calls to identifyPhysicalDrive, etc, will fail until the next reboot.
      dp.resetAdapterRaidConfig(adapterID, false /* factory defaults */);

      progress.setValue(progress.getMaximum());
      label.setText(JCRMUtil.getNLSString("scanExitInstruction"));
      doneButton.setEnabled(true);
      cancelButton.setEnabled(false);           

      doneButton.requestFocus();
   }         

   /**
   *  If the scanning for new drives should be terminated, then the appropriate
   *  events are sent indicating how many drives were found/removed.  This is 
   *  necessary so that the found/removed drives are reflected in the tree view.
   *  @param shouldScanEnd TRUE if the scanning for new/removed drives should be terminated; FALSE otherwise.
   *  @return TRUE if the scanning should be terminated (and appropriate events have been sent); FALSE otherwise.
   */
   private boolean endScan( boolean shouldScanEnd )
   {

      if (shouldScanEnd)
      {
         if (numAdded > 0)
         {
            Object[] args = { new Integer(numAdded), new Integer(adapter.getAdjustedID())};
            String eventKey = (numAdded == 1) ? "eventFoundNewReady" : "eventFoundNewReadys";
            String s = JCRMUtil.makeNLSString( eventKey, args);
            dp.getManagedSystem().clientGeneratedAlert(new ClientRaidEvent(dp.getServerName(),
                                                                                   Constants.INFORMATION_EVENT,
                                                                                   eventKey,
                                                                                   args,
                                                                                   s,
                                                                                   adapterID));
         }

         if (numRemoved > 0)
         {
            Object[] args = { new Integer(numRemoved), new Integer(adapterID+1)};
            String eventKey = (numRemoved) == 1 ? "eventFoundNewRemoved" : "eventFoundNewRemoveds";
            String s = JCRMUtil.makeNLSString( eventKey, args);
            dp.getManagedSystem().clientGeneratedAlert(new ClientRaidEvent(dp.getServerName(),
                                                                                   Constants.INFORMATION_EVENT,
                                                                                   eventKey,
                                                                                   args,
                                                                                   s,
                                                                                   adapterID));
         }

         // Always send an internal event so the gui gets refreshed. If the user replaces
         // a ready drive with another drive and does a scan, then no drives will be added or
         // removed, but the interface will need to be refreshed since the device information 
         // has changed.
         dp.getManagedSystem().clientGeneratedAlert(new ClientRaidEvent(dp.getServerName(),
                                                                        Constants.INTERNAL_EVENT,
                                                                        null, null, null, 0,
                                                                        false /*update overall status*/));
      }
      return shouldScanEnd;
   }

   /** This method, called at the end of the discovery process, scans for removed enclosure services
     * devices.  If, and only if, a channel contains no other devices, we need to attempt to remove
     * the enclosures on that channel.  Because of this additional requirement, we perform this task
     * here rather than in the main discovery loop.
     * @return true if enclosure devices were removed
     */
   private boolean removeEnclosureDevices()
   {
      boolean removed = false;
      Adapter adapter = null;

      // Rather than do lots of housekeeping during the discovery process, just get the updated config now
      RaidSystem system = dp.getConfigAll();

      // Ensure we got the config OK
      if (system == null)
         return false;

      // Loop through the adapters to find the right one
      for (Enumeration e = system.enumerateAdapters(); e.hasMoreElements();)
      {
         Adapter ad = (Adapter) e.nextElement();
         if (ad.getID() == adapterID)
         {
            adapter = ad;
            break;
         }
      }

      // Ensure we found the right adapter
      if (adapter == null)
         return false;

      // see if we need to abort
      if (endScan(stop))
         return false;

      // Loop through the channels
      for (Enumeration e2 = adapter.enumerateChannels(); e2.hasMoreElements();)
      {
         Channel channel = (Channel) e2.nextElement();

         if ((channelID != 99) && (channel.getID() != channelID)) // need specific channel and this ain't it
            continue;

         // see if this channel has *only* enclosures
         if ((channel.getNonEmptyDriveCount() == 0) && channel.containsAnySAFTEDevice())
         {
            for (Enumeration e3 = channel.getEnclosures(); e3.hasMoreElements();)
            {
               PhysicalDevice dev = (PhysicalDevice) e3.nextElement();

               // set each enclosure from SBY to EMP to SBY to effect removal
               StorRet ret = dp.setPhysDevState(new DeviceID(adapterID, dev.getChannelID(), dev.getDeviceID()), (short) Constants.EMP);
               if (ret.iReturnCode == 0)
               {
                  ret = dp.setPhysDevState(new DeviceID(adapterID, dev.getChannelID(), dev.getDeviceID()), (short) Constants.SBY);
                  if (ret.iReturnCode != 0)
                     removed = true;
               }

               // see if we need to abort
               if (endScan(stop))
                  return removed;

            } // enclosures loop
         } // channel valid test
      } // channel loop

      return removed;
   }

   void setupGUI()
   {
      // Setup the encompassing panel
      GridBagConstraints c = new GridBagConstraints();
      c.insets = new Insets(5,5,5,5);

      // Progress bar panel
      JPanel progressPanel = new JPanel();
      progressPanel.setBorder(new CompoundBorder(new EtchedBorder(), new EmptyBorder(10, 10, 10, 10)));
      progressPanel.setLayout(new GridLayout(5,1,5,5));

      progress = new JProgressBar();
      progress.setMaximum(99);
      progress.setMinimum(0);
      JPanel progressPane = new JPanel(new BorderLayout());
      progressPane.add(new JLabel(" 0% "), "West");
      progressPane.add(progress, "Center");
      progressPane.add(new JLabel(" 100% "), "East");
      progressPanel.add(progressPane);

      Box box = Box.createHorizontalBox();
      box.add(box.createHorizontalStrut(32));
      label = new JLabel(JCRMUtil.getNLSString("scanInstruction"));
      box.add(label);
      progressPanel.add(box);

      Box updatep1 = Box.createHorizontalBox();
      Box updatep2 = Box.createHorizontalBox();
      Object[] args = { new Integer(0)};
      String s1 = JCRMUtil.makeNLSString( "scanNumFound", args);
      String s2 = JCRMUtil.makeNLSString( "scanNumRemoved", args);
      ladd    = new JLabel(s1);
      lrem    = new JLabel(s2);
      updatep1.add(Box.createHorizontalStrut(32));
      updatep1.add(ladd);
      updatep2.add(Box.createHorizontalStrut(32));
      updatep2.add(lrem);
      progressPanel.add(updatep1);
      progressPanel.add(updatep2);

      progressPanel.setPreferredSize(new Dimension(progressPanel.getPreferredSize().width + 20, progressPanel.getPreferredSize().height));

      // Button Panel
      JPanel buttonPanel = new JPanel();

      cancelButton = new JButton(JCRMUtil.getNLSString("scanCancelButton"));
      cancelButton.setActionCommand("cancel");
      cancelButton.addActionListener(this);
      cancelButton.setMnemonic(JCRMUtil.getNLSString("cancelAltKey").charAt(0));

      doneButton = new JButton(JCRMUtil.getNLSString("scanDoneButton"));
      doneButton.setActionCommand("done");
      doneButton.addActionListener(this);
      doneButton.setMnemonic(JCRMUtil.getNLSString("doneAltKey").charAt(0));
      doneButton.setEnabled(false);

      buttonPanel.add(doneButton);
      buttonPanel.add(cancelButton);
      
      c.gridx = 0; c.gridy = 0; c.gridwidth = 3; c.gridheight = 3;
      add(progressPanel, c);

      c.gridx = 0; c.gridy = 5; c.gridwidth = 3; c.gridheight = 3;
      add(buttonPanel, c);

      this.addFocusListener(new FocusAdapter()
                           {
                              public void focusGained(FocusEvent evt)
                              {
                                 System.out.println("Focus gained");

                                 cancelButton.requestFocus();
                              }
                           });

      this.addKeyListener(new KeyAdapter()
                          {
                             public void keyReleased(KeyEvent e)
                             {
                                System.out.println("Key released " + e);

                                // Escape key is the same as CANCEL
                                if (e.getKeyCode() == KeyEvent.VK_ESCAPE )
                                {
                                   if (cancelButton.isEnabled() )
                                      cancelButton.doClick();
                                   else
                                      doneButton.doClick();
                                }
                             } 
                          });
   }

   public void actionPerformed(ActionEvent e)
   {
      if (e.getActionCommand().equals("cancel"))
      {
         stop = true;
         launch.endScanForNew();
      }
      else if (e.getActionCommand().equals("done"))
      {
         endScan(true);
         launch.endScanForNew();
      }
   }
}
